#	zgdchart.py
#
#	Change History:
#	0.3.4
#
#	29 Aug 2000	Graphs 2D and 3D chart styles
#			Fix bug in Edit Scale tab
#			Simplified form to instantiate class
#	28 Aug 2000	_chart is now thread-safe
#       18 May 2000	Urlencode encodes dates properly
#	12 Apr 2000	Fixes bugs introduced in 0.3.2
#			Improved code readability	
#

import Acquisition
import AccessControl
import OFS
import urllib
import string
import cStringIO
from Globals import HTMLFile, MessageDialog, Persistent
from Globals import package_home
from DocumentTemplate.DT_Util import *
from DocumentTemplate.DT_String import String
from maybe_lock import allocate_lock

#
# In this module, extend gdchart so
# that it can render itself
#
from gdchart import *

manage_addZGDChartForm=HTMLFile('addZGDChart',globals())
_chart_lock = allocate_lock()

def manage_addZGDChart(self,id,title, height=None, width=None, SQL=None, chart_type=None, option=[], REQUEST=None):
	"""
	Create a ZGDChart and install it in its parent Folder
	The argument 'self' will be bound to the parent Folder.
	"""

	handle=ZGDChart(id, title, height, width, SQL, chart_type, option)
	handle.id=id
	handle.title=title
	self._setObject(id,handle)
	if REQUEST is not None:
		return self.manage_main(self, REQUEST)

class ZGDChart(
	Acquisition.Implicit,
	Persistent,
	AccessControl.Role.RoleManager,
	OFS.SimpleItem.Item
	):
	'''
	A simple chart class which plots SQL data.
	The chart plots the first two columns of a SQL query result.

	This product uses gdchart0.94b, which in turn uses gd1.3
	The current version of gd 1.7.3 does not support gif creation due to the Unisys patent.

	Wrapper authored by <a href="mailto:teyc@bigfoot.com">Chui Tey</a>
	
	(c) Cognoware Australia P/L

	Software is provided without warranty.

	Credits:

	gdchart 	- Bruce Verdeaime   <A HREF="http://www.fred.net/brv/chart/">http://www.fred.net/brv/chart/</A>
	
	gdchart.py	- Michael Ray Steed <A HREF="http://members.xoom.com/_XMCM/msteed/software/gdchart.html"> http://members.xoom.com/_XMCM/msteed/software/gdchart.html </A>
	
	This product was created using mkproduct 0.6 and Zope Simple Skeleton v0.2,
	written by Sebastian L&uuml;hnsdorf (<a href="mailto:basti@beehive.de">basti@beehive.de</a>)

	For help, refer to the <A href="http://localhost/HelpSys/ObjectRef/ZGDChart/hs_main">Zope help system</A>
	'''

	name	=	'ZGDChart'
	id  	=	'ZGDChart'
	icon	=	'misc_/ZGDChart/Main'
	meta_type='ZGDChart'

	_properties = (
		{'id':'title','type':'string','mode':'w'},
		)

	manage_options=(
		{'label':'Properties','icon':icon,'action':'manage_main',  'target':'manage_main', 'help':('ZGDChart','ZGDChart_properties.stx')},
		{'label':'Colors'    ,'icon':icon,'action':'manage_mainColor','target':'manage_main'},
		{'label':'Scale'     ,'icon':icon,'action':'manage_mainScale','target':'manage_main'},
		{'label':'Security'  ,'icon':icon,'action':'manage_access','target':'manage_main'},
		{'label':'View'      ,'icon':icon,'action':'index_html',   'target':'manage_main'},
		)

	__ac_permissions__=(
		('View management screens',('manage_tabs','manage_main')),
		('Change permissions',('manage_access',)),
		('Change ZGDChart',('manage_edit','manage_editColor'),('Manager',)),
		)

	def home(self):
		"returns the directory where the product is located in"
		return package_home(globals())
	
	def version (self):
		"Return the version number of this product"		
		return open(self.home() + '/version.txt').read()
	
	def __init__(self,id='ZGDChart',title='ZGDChart', height=250, width=250, 
		SQL=None, chart_type='2D_Pie', option=[]):
		
		self.id 	    	= id
		self.title  	   	= title
		self.height 	   	= height
		self.width  	   	= width
		self.SQL    	   	= SQL
		self.content_type	= 'image/gif'
		self.tempfile		= self.home() + '/temp.gif'
			
		# Set the chart options
		self.bg_transparent	= 'bg_transparent' in option
		self.xaxis    		= 'xaxis' in option
		self.yaxis      	= 'yaxis' in option
		self.yaxis2		= 'yaxis2' in option
		self.label_line		= 'label_line' in option
		self.grid		= 'grid' in option
		self.border		= 'border' in option
		self.thumbnail		= 'thumbnail' in option
		self.chart_type 	= chart_type

		# Set the chart color
		self.bg_color		=0xFFFFFF
		self.title_color	=0x446688
		self.edge_color		=0x000000
		self.grid_color		=0x446688
		self.plot_color		=0x446688
		self.xlabel_color	=0x446688
		self.ylabel_color	=0x446688
		self.ylabel2_color	=0x446644
		self.xtitle_color	=0x446688
		self.ytitle_color	=0x446688
		self.ytitle2_color	=0x446644

		# ---------------------------------
		# Declare these variables because
		# we don't want to break compatibility
		# when these evetually get implemented
		
		# Set chart text
		self.charttitle 	= title
		self.xtitle		= ''
		self.ytitle		= ''
		self.ytitle2		= ''

		# Set chart text font sizes
		self.title_font		= GDC_LARGE
		self.xtitle_font	= GDC_SMALL
		self.ytitle_font	= GDC_SMALL
		self.ytitle2_font	= GDC_SMALL
		self.annotation_font	= GDC_SMALL

		# Set miscellaneous items
		self.bar_width			=	10
		self.explode 			=	[0,0,0]
		self.ext_color			=	[0,0,0]
		self.ext_vol_color 		=	[0,0,0]
		self.generate_gif		=	1
		self.hard_graphheight		=	300
		self.hard_graphwidth		= 	400
		self.hard_size			=	0
		self.hard_xorig			=	0
		self.hard_yorig			=	0
		self.hlc_cap_width		=	0
		self.hlc_style			=	GDC_HLC_CLOSE_CONNECTED
		self.hold_img			=	0
		self.label_dist			=	20
		self.missing			= 	[]
		self.other_threshold		= 	10
		self.percent_labels		=	GDCPIE_PCT_NONE
		self.pie_color			= 	[0xAACCEE, 0x88AACC, 0x6688AA]
		self.thumbval			=	0
		self.zeroshelf			=	0
		
		# Set axis
		self.requested_ymax		=	None
		self.requested_ymin		=	None
		self.requested_yinterval	=	None
		self.xlabel_spacing		=	None
		self.ylabel_density		=	None
		self.ylabel_fmt			=	''
		self.ylabel2_fmt		=	''
		
		# Set three dimensional properties		
		self.threed_angle		=	0
		self.threed_depth		=	0
						
	manage_main=HTMLFile('manage/edit', globals())
	manage_mainColor = HTMLFile('manage/editColor', globals())
	manage_mainScale = HTMLFile('manage/editScale', globals())
    
	def manage_edit(self,title, xtitle, ytitle, width, height, SQL, chart_type, option=[], REQUEST=None):
		"""edit ZGDChart"""
		self.title      = title
		self.xtitle     = xtitle
		self.ytitle	= ytitle
		self.height     = height
		self.width      = width
		self.SQL        = SQL
		self.chart_type = chart_type
		self.bg_transparent = 'bg_transparent' in option
		self.label_line     = 'label_line' in option
		self.thumbnail 	    = 'thumbnail' in option
		self.border     = 'border' in option
		self.grid	    = 'grid' in option
		self.xaxis      = 'xaxis' in option
		self.yaxis      = 'yaxis' in option
		self.yaxis2     = 'yaxis2' in option

		if REQUEST is not None:
			return MessageDialog(
				title='Edited',
				message='<strong>%s</strong> has been edited.' % self.id,
				action = './manage_main'
				)

	def manage_editColor(self, bg_color, edge_color, grid_color, plot_color,
						 title_color, xtitle_color, ytitle_color, ytitle2_color,
						 xlabel_color, ylabel_color, ylabel2_color,
						 REQUEST=None):
		""" edit ZGDChart colors
			all values to be passed as strings in hex, eg..
			manage_editColor(bg_color = "0xFFFFFF")
		"""
		self.bg_color   	= string.atoi(bg_color,   16)
		self.edge_color 	= string.atoi(edge_color, 16)
		self.grid_color 	= string.atoi(grid_color, 16)
		self.plot_color 	= string.atoi(plot_color, 16)
		self.title_color	= string.atoi(title_color, 16)
		self.xtitle_color 	= string.atoi(xtitle_color, 16)
		self.ytitle_color 	= string.atoi(ytitle_color, 16)
		self.ytitle2_color 	= string.atoi(ytitle2_color, 16)
		self.xlabel_color	= string.atoi(xlabel_color, 16)
		self.ylabel_color 	= string.atoi(ylabel_color, 16)
		self.ylabel2_color	= string.atoi(ylabel2_color, 16)
		
		if REQUEST is not None:
			return MessageDialog(
				title='Edited',
				message='<strong>%s</strong> has been edited.' % self.id,
				action = './manage_main'
				)

	def manage_editScale(self,
				 requested_ymax=None, requested_ymin=None, requested_yinterval=None,
				 xlabel_spacing=0, ylabel_density=80,
				 ylabel_fmt='', ylabel2_fmt='', REQUEST=None):
		""" edit ZGDChart scale properties """		
		self.xlabel_spacing		=	xlabel_spacing

		self.requested_ymax		=	requested_ymax		
		self.requested_ymin		=	requested_ymin
		self.requested_yinterval	=	requested_yinterval
		self.ylabel_density		=	ylabel_density
		self.ylabel_fmt			=	ylabel_fmt
		self.ylabel2_fmt		=	ylabel2_fmt

		if REQUEST is not None:
			return MessageDialog(
				title='Edited',
				message='<strong>%s</strong> has been edited.' % self.id,
				action = './manage_main'
				)
		    
	def __call__(self, mapping={}, **kwargs):
		''' Just another version of __str__ '''
		# Not sure if this is going to block other inherited methods

		# combine mapping and kwargs into a single dictionary
		dict = {}
		for key,value in mapping.items():
			dict[key]=value
		for key,value in kwargs.items():
			dict[key]=value
		
		# roll our own urlencode because we want
		# / to be encoded
		# query = urllib.urlencode(kwargs)
		query = self.urlencode(dict)
		return '<IMG SRC="%s?%s" ALT="%s">' % (self.id, query, self.title)

	def urlencode(self, dict):
		l = []
		for k, v in dict.items():
			k = urllib.quote_plus(str(k))
			v = urllib.quote_plus(str(v),'')
			l.append(k + '=' + v)
		return string.join(l, '&')

	# following routines are inspired by Arpad Kiss's chart product
	def __str__(self,REQUEST=None,RESPONSE=None):
		"""
		Returns a string representation of the object <IMG SRC=xxx>
		"""
		return '<IMG SRC="%s" ALT="%s">' % (self.id, self.title)
	
	def index_html(self,REQUEST=None,RESPONSE=None):
		"""
		This is the default method. It calls the right renderer: Line,Column or Pie
		"""
		return eval("self.%s(REQUEST,RESPONSE)" % self.chart_type)

	#
	#	Methods of a ZGDChart instance
	#	returns a chart
	#
	#	The first few methods area inconsistent
	#	in their implementations of 2D and 3D
	#	charts. Leave them unchanged to preserve
	#	backward compatibility
	#
	def Area(self, REQUEST=None, RESPONSE=None):
		"""
		Renders dataset as a bar chart
		Deprecated
		"""
		return self._chart(GDC_AREA, RESPONSE)

	def Bar(self, REQUEST=None, RESPONSE=None):
		"""
		Renders dataset as a bar chart
		Deprecated
		"""
		return self._chart(GDC_BAR, RESPONSE)

	def Line(self,REQUEST=None,RESPONSE=None):
		"""
		Renders dataset as a line chart
		Deprecated
		"""
		return self._chart(GDC_LINE, RESPONSE)

	def Pie(self,REQUEST=None,RESPONSE=None):
		"""
		Renders dataset as a pie chart
		Deprecated
		"""
		return self._chart(GDC_3DPIE, RESPONSE)

	#
	#	Implements the new naming API
	#

	def Area_2D(self, REQUEST=None, RESPONSE=None):
		"""
		Renders dataset as a bar chart
		"""
		return self._chart(GDC_AREA, RESPONSE)

	def Bar_2D(self, REQUEST=None, RESPONSE=None):
		"""
		Renders dataset as a bar chart
		"""
		return self._chart(GDC_BAR, RESPONSE)

	def Line_2D(self,REQUEST=None,RESPONSE=None):
		"""
		Renders dataset as a line chart
		"""
		return self._chart(GDC_LINE, RESPONSE)

	def Pie_2D(self, REQUEST=None, RESPONSE=None):
		"""
		Renders dataset as a 2D pie chart
		"""
		return self._chart(GDC_2DPIE, RESPONSE)
			
	def Area_3D(self, REQUEST=None, RESPONSE=None):
		"""
		Renders dataset as a 3D area chart
		"""
		return self._chart(GDC_3DAREA, RESPONSE)
			
	def Bar_3D(self, REQUEST=None, RESPONSE=None):
		"""
		Renders dataset as a 3D bar chart
		"""
		return self._chart(GDC_3DBAR, RESPONSE)

	def Line_3D(self, REQUEST=None, RESPONSE=None):
		"""
		Renders dataset as a 3D line chart
		"""
		return self._chart(GDC_3DLINE, RESPONSE)
			
	def Pie_3D(self, REQUEST=None, RESPONSE=None):
		"""
		Renders dataset as a 3D pie chart
		"""
		return self._chart(GDC_3DPIE, RESPONSE)

	#	XXX To do: Hack chartdata to
	#	           handle hi lo close charts
	#
	def Hi_Lo_Close_2D(self, REQUEST=None, RESPONSE=None):
		"""
		Renders dataset as a hi lo close chart
		"""
		return self._chart(GDC_HI_LO_CLOSE, RESPONSE)		
	
	def Hi_Lo_Close_3D(self, REQUEST=None, RESPONSE=None):
		"""
		Renders dataset as a 3D combo of Hi_Lo_Close
		"""
		return self._chart(GDC_3DHILOCLOSE, RESPONSE)

	#	XXX To do: Hack chartdata to
	#	           handle combo charts
	#
	def Combo_Line_Area_2D(self, REQUEST=None, RESPONSE=None):
		"""
		Renders dataset as a combo of line and area chart
		"""
		return self._chart(GDC_COMBO_LINE_AREA, RESPONSE)

	def Combo_Line_Bar_2D(self, REQUEST=None, RESPONSE=None):
		"""
		Renders dataset as a combo of line and bar chart
		"""
		return self._chart(GDC_COMBO_LINE_BAR, RESPONSE)

	def Combo_HLC_Area_2D(self, REQUEST=None, RESPONSE=None):
		"""
		Renders dataset as a combo of hi lo close and area chart
		"""
		return self._chart(GDC_COMBO_HLC_AREA, RESPONSE)

	def Combo_HLC_Bar_2D(self, REQUEST=None, RESPONSE=None):
		"""
		Renders dataset as a combo of hi lo close and bar chart
		"""
		return self._chart(GDC_COMBO_HLC_BAR, RESPONSE)

	def Combo_Line_Area_3D(self, REQUEST=None, RESPONSE=None):
		"""
		Renders dataset as a 3D combo of line and area chart
		"""
		return self._chart(GDC_3DCOMBO_LINE_AREA, RESPONSE)
			
	def Combo_Line_Bar_3D(self, REQUEST=None, RESPONSE=None):
		"""
		Renders dataset as a 3D combo of line and area bar
		"""
		return self._chart(GDC_3DCOMBO_LINE_BAR, RESPONSE)
			
	def Combo_HLC_Area_3D(self, REQUEST=None, RESPONSE=None):
		"""
		Renders dataset as a 3D combo of HLC and area chart
		"""
		return self._chart(GDC_3DCOMBO_HLC_AREA, RESPONSE)
			
	def Combo_HLC_Bar_3D(self, REQUEST=None, RESPONSE=None):
		"""
		Renders dataset as a 3D combo of HLC and area bar
		"""
		return self._chart(GDC_3DCOMBO_HLC_BAR, RESPONSE)
			
	def _chart(self, GDC_TYPE, RESPONSE=None):
		"""
		Renders dataset in the format requested
		"""
		try:
			_chart_lock.acquire()
			self._setOption()
			self._setColor()
			self._setScale()
			labels, data = self._chartdata()
			# tempfile = cStringIO.StringIO(), requires GDChart 0.4
			chart(GDC_TYPE, (self.width, self.height), self.tempfile, labels, data)
			if RESPONSE: RESPONSE['content-type']=self.content_type
			# output = tempfile.getvalue() 
			output = open(self.tempfile,'rb').read()
		finally:
			_chart_lock.release()			
		return  output
		
	def _setOption(self):
		option(title = self.title)
		option(xtitle = self.xtitle)
		option(ytitle = self.ytitle)
		option(bg_transparent = self.bg_transparent)
		option(grid = self.grid)
		option(border = self.border)
		option(label_line= self.label_line)
		option(thumbnail = self.thumbnail)
		option(xaxis = self.xaxis)
		option(yaxis = self.yaxis)
		option(yaxis2= self.yaxis2)
		option(set_color = (0xaabbcc, 0xffbbcc, 0xaabb55, 0x3388aa, 0xaa8833, 0x88aa33))	
		option(pie_color = (0xaabbcc, 0xffbbcc, 0xaabb55, 0x3388aa, 0xaa8833, 0x88aa33))	
	
	def _setColor(self):
		option(bg_color  	= self.bg_color)
		option(grid_color	= self.grid_color)
		option(plot_color	= self.plot_color)
		option(edge_color 	= self.edge_color)
		option(title_color	= self.title_color)
		option(xtitle_color 	= self.xtitle_color)
		option(ytitle_color 	= self.ytitle_color)
		option(ytitle2_color 	= self.ytitle2_color)
		option(xlabel_color 	= self.xlabel_color)
		option(ylabel_color 	= self.ylabel_color)
		option(ylabel2_color 	= self.ylabel2_color)

	def _setScale(self):
		#
		#  the only possible way to reset the scaling options
		#  to default values is to use a patched gdchart.pyd,
		#  and pass None as a parameter. self.requested_ymax and
		#  other attributes in the following block may be a None
		#
		try:
			option(requested_ymax =	self.requested_ymax)
		        option(requested_ymin =	self.requested_ymin)
		        option(requested_yinterval = self.requested_yinterval)
		        option(xlabel_spacing =	self.xlabel_spacing)
		        option(ylabel_density =	self.ylabel_density)
		except error:
			#
			#  probably haven't got a patched gdchart.pyd
			#  will try to run gdchart using previous scale settings
			#  it will be ugly but at least won't give an error page
			#
			#  XXX log this to a file?
			#
			print "GDChart setscale error"
			pass;
			
		option(ylabel_fmt	=	self.ylabel_fmt )
		option(ylabel2_fmt	=	self.ylabel2_fmt)

	def _chartdata(self):
		"""
		return the first and second columns of sql query
    		results as a pair of tuples
	        """
		rows = getattr(self, self.SQL)()
		labels= []
		data =  []
		for row in rows:
			# each row is a list of fields
 			fields = row
			#
			# Only insert rows with zero values if
			# it is not a pie chart
			#
			if self.chart_type!='Pie' or float(fields[1])>0:
				#
				#	Coerce the first field into a string
				#	and second field into a float
				#
        			labels.append( str  (fields[0]))
        			data.append  ( float(fields[1]))
    		#
    		# convert lists to tuples
    		# apparently, gdchart.chart() accepts lists
    		# as well, but I better check with
    		# M Steed first.
    		#
    		return tuple(labels), tuple(data)
		
	def _labels(self):
		"""
		return the first column sql query results as a tuple
	        """
		rows = getattr(self, self.SQL)()
		tmp=()
		for row in rows:
		   tmp = tmp + (row[0],)
		return tmp
		
	def _data(self):
		rows = getattr(self, self.SQL)()
		tmp=()
		for row in rows:
		    tmp = tmp + (row[1],)
		return tmp

